from math import ceil
from typing import Generator, List, Union

import numpy as np
import torch
from oneliner_utils import read_jsonl
from torch import Tensor, einsum
from torch.nn.functional import normalize
from tqdm import tqdm
from transformers import AutoConfig, AutoModel, AutoTokenizer

from ..paths import embeddings_folder_path, encoder_state_path

pbar_kwargs = dict(position=0, dynamic_ncols=True, mininterval=1.0)


def count_lines(path: str):
    """Counts the number of lines in a file."""
    return sum(1 for _ in open(path))


def generate_batch(docs: Union[List, Generator], batch_size: int) -> Generator:
    texts = []

    for doc in docs:
        texts.append(doc["text"])

        if len(texts) == batch_size:
            yield texts
            texts = []

    if texts:
        yield texts


class Encoder:
    def __init__(
        self,
        index_name: str = "new-index",
        model: str = "sentence-transformers/all-MiniLM-L6-v2",
        normalize: bool = True,
        return_numpy: bool = True,
        max_length: int = 128,
        device: str = "cpu",
    ):
        self.index_name = index_name
        self.model = model
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.encoder = AutoModel.from_pretrained(model).to(device).eval()
        self.embedding_dim = AutoConfig.from_pretrained(model).hidden_size
        self.max_length = max_length
        self.normalize = normalize
        self.return_numpy = return_numpy
        self.device = device
        self.tokenizer_kwargs = {
            "padding": True,
            "truncation": True,
            "max_length": self.max_length,
            "return_tensors": "pt",
        }

    def save(self):
        state = dict(
            index_name=self.index_name,
            model=self.model,
            normalize=self.normalize,
            return_numpy=self.return_numpy,
            max_length=self.max_length,
            device=self.device,
        )
        np.save(encoder_state_path(self.index_name), state)

    @staticmethod
    def load(index_name: str, device: str = None):
        state = np.load(encoder_state_path(index_name), allow_pickle=True)[()]
        if device is not None:
            state["device"] = device
        return Encoder(**state)

    def change_device(self, device: str = "cpu"):
        self.device = device
        self.encoder.to(device)

    def tokenize(self, texts: List[str]):
        tokens = self.tokenizer(texts, **self.tokenizer_kwargs)
        return {k: v.to(self.device) for k, v in tokens.items()}

    def mean_pooling(self, embeddings: Tensor, mask: Tensor) -> Tensor:
        numerators = einsum("xyz,xy->xyz", embeddings, mask).sum(dim=1)
        denominators = torch.clamp(mask.sum(dim=1), min=1e-9)
        return einsum("xz,x->xz", numerators, 1 / denominators)

    def __call__(
        self,
        x: Union[str, List[str]],
        batch_size: int = 32,
        show_progress: bool = True,
    ):
        if isinstance(x, str):
            return self.encode(x)
        else:
            return self.bencode(x, batch_size=batch_size, show_progress=show_progress)

    def encode(self, text: str):
        return self.bencode([text], batch_size=1, show_progress=False)[0]

    def bencode(
        self,
        texts: List[str],
        batch_size: int = 32,
        show_progress: bool = True,
    ):
        embeddings = []

        pbar = tqdm(
            total=len(texts),
            desc="Generating embeddings",
            disable=not show_progress,
            **pbar_kwargs,
        )

        for i in range(ceil(len(texts) / batch_size)):
            start, stop = i * batch_size, (i + 1) * batch_size
            tokens = self.tokenize(texts[start:stop])

            with torch.no_grad():
                emb = self.encoder(**tokens).last_hidden_state
                emb = self.mean_pooling(emb, tokens["attention_mask"])
                if self.normalize:
                    emb = normalize(emb, dim=-1)

            embeddings.append(emb)

            pbar.update(stop - start)
        pbar.close()

        embeddings = torch.cat(embeddings)

        if self.return_numpy:
            embeddings = embeddings.detach().cpu().numpy()

        return embeddings

    def encode_collection(
        self,
        path: str,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        n_docs = count_lines(path)
        collection = read_jsonl(path, callback=callback, generator=True)

        reservoir = np.empty((1_000_000, self.embedding_dim), dtype=np.float32)
        reservoir_n = 0
        offset = 0

        pbar = tqdm(
            total=n_docs,
            desc="Embedding documents",
            disable=not show_progress,
            **pbar_kwargs,
        )

        for texts in generate_batch(collection, batch_size):
            # Compute embeddings -----------------------------------------------
            embeddings = self.bencode(texts, batch_size=len(texts), show_progress=False)

            # Compute new offset -----------------------------------------------
            new_offset = offset + len(embeddings)

            if new_offset >= len(reservoir):
                np.save(
                    embeddings_folder_path(self.index_name)
                    / f"chunk_{reservoir_n}.npy",
                    reservoir[:offset],
                )
                reservoir = np.empty((1_000_000, self.embedding_dim), dtype=np.float32)
                reservoir_n += 1
                offset = 0
                new_offset = len(embeddings)

            # Save embeddings in the reservoir ---------------------------------
            reservoir[offset:new_offset] = embeddings

            # Update offeset ---------------------------------------------------
            offset = new_offset

            pbar.update(len(embeddings))

        if offset < len(reservoir):
            np.save(
                embeddings_folder_path(self.index_name) / f"chunk_{reservoir_n}.npy",
                reservoir[:offset],
            )
            reservoir = []

        assert len(reservoir) == 0, "Reservoir is not empty."



import os
from typing import Dict, Iterable, List, Set, Union

import numba as nb
import numpy as np
import orjson
from numba.typed import List as TypedList
from oneliner_utils import create_path, read_jsonl
from tqdm import tqdm

from ..autotune import tune_bm25
from ..base_retriever import BaseRetriever
from ..paths import docs_path, sr_state_path
from .build_inverted_index import build_inverted_index
from .preprocessing import (
    get_stemmer,
    get_stopwords,
    get_tokenizer,
    preprocessing,
    preprocessing_multi,
)
from .sparse_retrieval_models.bm25 import bm25, bm25_multi
from .sparse_retrieval_models.tf_idf import tf_idf, tf_idf_multi


class SparseRetriever(BaseRetriever):
    def __init__(
        self,
        index_name: str = "new-index",
        model: str = "bm25",
        min_df: int = 1,
        tokenizer: Union[str, callable] = "whitespace",
        stemmer: Union[str, callable] = "english",
        stopwords: Union[str, List[str], Set[str]] = "english",
        do_lowercasing: bool = True,
        do_ampersand_normalization: bool = True,
        do_special_chars_normalization: bool = True,
        do_acronyms_normalization: bool = True,
        do_punctuation_removal: bool = True,
        hyperparams: dict = None,
    ):
        """The Sparse Retriever is a traditional searcher based on lexical matching. It supports BM25, the retrieval model used by major search engines libraries, such as Lucene and Elasticsearch. retriv also implements the classic relevance model TF-IDF for educational purposes.

        Args:
            index_name (str, optional): [retriv](https://github.com/AmenRa/retriv) will use `index_name` as the identifier of your index. Defaults to "new-index".

            model (str, optional): defines the retrieval model to use for searching (`bm25` or `tf-idf`). Defaults to "bm25".

            min_df (int, optional): terms that appear in less than `min_df` documents will be ignored. If integer, the parameter indicates the absolute count. If float, it represents a proportion of documents. Defaults to 1.

            tokenizer (Union[str, callable], optional): [tokenizer](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to use during preprocessing. You can pass a custom callable tokenizer or disable tokenization by setting the parameter to `None`. Defaults to "whitespace".

            stemmer (Union[str, callable], optional): [stemmer](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to use during preprocessing. You can pass a custom callable stemmer or disable stemming setting the parameter to `None`. Defaults to "english".

            stopwords (Union[str, List[str], Set[str]], optional): [stopwords](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to remove during preprocessing. You can pass a custom stop-word list or disable stop-words removal by setting the parameter to `None`. Defaults to "english".

            do_lowercasing (bool, optional): whether or not to lowercase texts. Defaults to True.

            do_ampersand_normalization (bool, optional): whether to convert `&` in `and` during pre-processing. Defaults to True.

            do_special_chars_normalization (bool, optional): whether to remove special characters for letters, e.g., `übermensch` → `ubermensch`. Defaults to True.

            do_acronyms_normalization (bool, optional): whether to remove full stop symbols from acronyms without splitting them in multiple words, e.g., `P.C.I.` → `PCI`. Defaults to True.

            do_punctuation_removal (bool, optional): whether to remove punctuation. Defaults to True.

            hyperparams (dict, optional): Retrieval model hyperparams. If `None`, it is automatically set to `{b: 0.75, k1: 1.2}`. Defaults to None.
        """

        assert model.lower() in {"bm25", "tf-idf"}
        assert min_df > 0, "`min_df` must be greater than zero."
        self.init_args = {
            "model": model.lower(),
            "min_df": min_df,
            "index_name": index_name,
            "do_lowercasing": do_lowercasing,
            "do_ampersand_normalization": do_ampersand_normalization,
            "do_special_chars_normalization": do_special_chars_normalization,
            "do_acronyms_normalization": do_acronyms_normalization,
            "do_punctuation_removal": do_punctuation_removal,
            "tokenizer": tokenizer,
            "stemmer": stemmer,
            "stopwords": stopwords,
        }

        self.model = model.lower()
        self.min_df = min_df
        self.index_name = index_name

        self.do_lowercasing = do_lowercasing
        self.do_ampersand_normalization = do_ampersand_normalization
        self.do_special_chars_normalization = do_special_chars_normalization
        self.do_acronyms_normalization = do_acronyms_normalization
        self.do_punctuation_removal = do_punctuation_removal

        self.tokenizer = get_tokenizer(tokenizer)
        self.stemmer = get_stemmer(stemmer)
        self.stopwords = [self.stemmer(sw) for sw in get_stopwords(stopwords)]

        self.id_mapping = None
        self.inverted_index = None
        self.vocabulary = None
        self.doc_count = None
        self.doc_lens = None
        self.avg_doc_len = None
        self.relative_doc_lens = None
        self.doc_index = None

        self.preprocessing_kwargs = {
            "tokenizer": self.tokenizer,
            "stemmer": self.stemmer,
            "stopwords": self.stopwords,
            "do_lowercasing": self.do_lowercasing,
            "do_ampersand_normalization": self.do_ampersand_normalization,
            "do_special_chars_normalization": self.do_special_chars_normalization,
            "do_acronyms_normalization": self.do_acronyms_normalization,
            "do_punctuation_removal": self.do_punctuation_removal,
        }

        self.preprocessing_pipe = preprocessing_multi(**self.preprocessing_kwargs)

        self.hyperparams = dict(b=0.75, k1=1.2) if hyperparams is None else hyperparams

    def save(self) -> None:
        """Save the state of the retriever to be able to restore it later."""

        state = {
            "init_args": self.init_args,
            "id_mapping": self.id_mapping,
            "doc_count": self.doc_count,
            "inverted_index": self.inverted_index,
            "vocabulary": self.vocabulary,
            "doc_lens": self.doc_lens,
            "relative_doc_lens": self.relative_doc_lens,
            "hyperparams": self.hyperparams,
        }

        np.savez_compressed(sr_state_path(self.index_name), state=state)

    @staticmethod
    def load(index_name: str = "new-index"):
        """Load a retriever and its index.

        Args:
            index_name (str, optional): Name of the index. Defaults to "new-index".

        Returns:
            SparseRetriever: Sparse Retriever.
        """

        state = np.load(sr_state_path(index_name), allow_pickle=True)["state"][()]

        se = SparseRetriever(**state["init_args"])
        se.initialize_doc_index()
        se.id_mapping = state["id_mapping"]
        se.doc_count = state["doc_count"]
        se.inverted_index = state["inverted_index"]
        se.vocabulary = set(se.inverted_index)
        se.doc_lens = state["doc_lens"]
        se.relative_doc_lens = state["relative_doc_lens"]
        se.hyperparams = state["hyperparams"]

        state = {
            "init_args": se.init_args,
            "id_mapping": se.id_mapping,
            "doc_count": se.doc_count,
            "inverted_index": se.inverted_index,
            "vocabulary": se.vocabulary,
            "doc_lens": se.doc_lens,
            "relative_doc_lens": se.relative_doc_lens,
            "hyperparams": se.hyperparams,
        }

        return se

    def index_aux(self, show_progress: bool = True):
        """Internal usage."""
        collection = read_jsonl(
            docs_path(self.index_name),
            generator=True,
            callback=lambda x: x["text"],
        )

        # Preprocessing --------------------------------------------------------
        collection = self.preprocessing_pipe(collection, generator=True)

        # Inverted index -------------------------------------------------------
        (
            self.inverted_index,
            self.doc_lens,
            self.relative_doc_lens,
        ) = build_inverted_index(
            collection=collection,
            n_docs=self.doc_count,
            min_df=self.min_df,
            show_progress=show_progress,
        )
        self.avg_doc_len = np.mean(self.doc_lens, dtype=np.float32)
        self.vocabulary = set(self.inverted_index)

    def index(
        self,
        collection: Iterable,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index a given collection of documents.

        Args:
            collection (Iterable): collection of documents to index.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            SparseRetriever: Sparse Retriever.
        """

        self.save_collection(collection, callback)
        self.initialize_doc_index()
        self.initialize_id_mapping()
        self.doc_count = len(self.id_mapping)
        self.index_aux(show_progress)
        self.save()
        return self

    def index_file(
        self, path: str, callback: callable = None, show_progress: bool = True
    ):
        """Index the collection contained in a given file.

        Args:
            path (str): path of file containing the collection to index.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            SparseRetriever: Sparse Retriever
        """

        collection = self.collection_generator(path=path, callback=callback)
        return self.index(collection=collection, show_progress=show_progress)

    # SEARCH ===================================================================
    def query_preprocessing(self, query: str) -> List[str]:
        """Internal usage."""
        return preprocessing(query, **self.preprocessing_kwargs)

    def get_term_doc_freqs(self, query_terms: List[str]) -> nb.types.List:
        """Internal usage."""
        return TypedList([self.inverted_index[t]["tfs"] for t in query_terms])

    def get_doc_ids(self, query_terms: List[str]) -> nb.types.List:
        """Internal usage."""
        return TypedList([self.inverted_index[t]["doc_ids"] for t in query_terms])

    def search(self, query: str, return_docs: bool = True, cutoff: int = 100) -> List:
        """Standard search functionality.

        Args:
            query (str): what to search for.

            return_docs (bool, optional): wether to return the texts of the documents. Defaults to True.

            cutoff (int, optional): number of results to return. Defaults to 100.

        Returns:
            List: results.
        """

        query_terms = self.query_preprocessing(query)
        if not query_terms:
            return {}
        query_terms = [t for t in query_terms if t in self.vocabulary]
        if not query_terms:
            return {}

        doc_ids = self.get_doc_ids(query_terms)
        term_doc_freqs = self.get_term_doc_freqs(query_terms)

        if self.model == "bm25":
            unique_doc_ids, scores = bm25(
                term_doc_freqs=term_doc_freqs,
                doc_ids=doc_ids,
                relative_doc_lens=self.relative_doc_lens,
                doc_count=self.doc_count,
                cutoff=cutoff,
                **self.hyperparams,
            )
        elif self.model == "tf-idf":
            unique_doc_ids, scores = tf_idf(
                term_doc_freqs=term_doc_freqs,
                doc_ids=doc_ids,
                doc_lens=self.doc_lens,
                cutoff=cutoff,
            )
        else:
            raise NotImplementedError()

        unique_doc_ids = self.map_internal_ids_to_original_ids(unique_doc_ids)

        if not return_docs:
            return dict(zip(unique_doc_ids, scores))

        return self.prepare_results(unique_doc_ids, scores)

    def msearch(self, queries: List[Dict[str, str]], cutoff: int = 100) -> Dict:
        """Compute results for multiple queries at once.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

        Returns:
            Dict: results.
        """

        term_doc_freqs = TypedList()
        doc_ids = TypedList()
        q_ids = []
        no_results_q_ids = []

        for q in queries:
            q_id, query = q["id"], q["text"]
            query_terms = self.query_preprocessing(query)
            query_terms = [t for t in query_terms if t in self.vocabulary]
            if not query_terms:
                no_results_q_ids.append(q_id)
                continue

            if all(t not in self.inverted_index for t in query_terms):
                no_results_q_ids.append(q_id)
                continue

            q_ids.append(q_id)
            term_doc_freqs.append(self.get_term_doc_freqs(query_terms))
            doc_ids.append(self.get_doc_ids(query_terms))

        if not q_ids:
            return {q_id: {} for q_id in [q["id"] for q in queries]}

        if self.model == "bm25":
            unique_doc_ids, scores = bm25_multi(
                term_doc_freqs=term_doc_freqs,
                doc_ids=doc_ids,
                relative_doc_lens=self.relative_doc_lens,
                doc_count=self.doc_count,
                cutoff=cutoff,
                **self.hyperparams,
            )
        elif self.model == "tf-idf":
            unique_doc_ids, scores = tf_idf_multi(
                term_doc_freqs=term_doc_freqs,
                doc_ids=doc_ids,
                doc_lens=self.doc_lens,
                cutoff=cutoff,
            )
        else:
            raise NotImplementedError()

        unique_doc_ids = [
            self.map_internal_ids_to_original_ids(_unique_doc_ids)
            for _unique_doc_ids in unique_doc_ids
        ]

        results = {
            q: dict(zip(unique_doc_ids[i], scores[i])) for i, q in enumerate(q_ids)
        }

        for q_id in no_results_q_ids:
            results[q_id] = {}

        # Order as queries
        return {q_id: results[q_id] for q_id in [q["id"] for q in queries]}

    def bsearch(
        self,
        queries: List[Dict[str, str]],
        cutoff: int = 100,
        batch_size: int = 1_000,
        show_progress: bool = True,
        qrels: Dict[str, Dict[str, float]] = None,
        path: str = None,
    ) -> Dict:
        """Batch-Search is similar to Multi-Search but automatically generates batches of queries to evaluate and allows dynamic writing of the search results to disk in [JSONl](https://jsonlines.org) format. bsearch is handy for computing results for hundreds of thousands or even millions of queries without hogging your RAM.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

            batch_size (int, optional): number of query to perform simultaneously. Defaults to 1_000.

            show_progress (bool, optional): whether to show a progress bar for the search process. Defaults to True.

            qrels (Dict[str, Dict[str, float]], optional): query relevance judgements for the queries. Defaults to None.

            path (str, optional): where to save the results. Defaults to None.

        Returns:
            Dict: results.
        """

        batches = [
            queries[i : i + batch_size] for i in range(0, len(queries), batch_size)
        ]

        results = {}

        pbar = tqdm(
            total=len(queries),
            disable=not show_progress,
            desc="Batch search",
            dynamic_ncols=True,
            mininterval=0.5,
        )

        if path is None:
            for batch in batches:
                new_results = self.msearch(queries=batch, cutoff=cutoff)
                results = {**results, **new_results}
                pbar.update(min(batch_size, len(batch)))
        else:
            path = create_path(path)
            path.parent.mkdir(parents=True, exist_ok=True)

            with open(path, "wb") as f:
                for batch in batches:
                    new_results = self.msearch(queries=batch, cutoff=cutoff)

                    for i, (k, v) in enumerate(new_results.items()):
                        x = {
                            "id": k,
                            "text": batch[i]["text"],
                            f"{self.model}_doc_ids": list(v.keys()),
                            f"{self.model}_scores": [
                                float(s) for s in list(v.values())
                            ],
                        }
                        if qrels is not None:
                            x["rel_doc_ids"] = list(qrels[k].keys())
                            x["rel_scores"] = list(qrels[k].values())
                        f.write(orjson.dumps(x) + "\n".encode())

                    pbar.update(min(batch_size, len(batch)))

        return results

    def autotune(
        self,
        queries: List[Dict[str, str]],
        qrels: Dict[str, Dict[str, float]],
        metric: str = "ndcg",
        n_trials: int = 100,
        cutoff: int = 100,
    ):
        """Use the AutoTune function to tune [BM25](https://en.wikipedia.org/wiki/Okapi_BM25) parameters w.r.t. your document collection and queries.
        All metrics supported by [ranx](https://github.com/AmenRa/ranx) are supported by the `autotune` function. At the of the process, the best parameter configuration is automatically applied to the `SparseRetriever` instance and saved to disk. You can inspect the current configuration by printing `sr.hyperparams`.

        Args:
            queries (List[Dict[str, str]]): queries to use for the optimization process.

            qrels (Dict[str, Dict[str, float]]): query relevance judgements for the queries.

            metric (str, optional): metric to optimize for. Defaults to "ndcg".

            n_trials (int, optional): number of configuration to evaluate. Defaults to 100.

            cutoff (int, optional): number of results to consider for the optimization process. Defaults to 100.
        """

        hyperparams = tune_bm25(
            queries=queries,
            qrels=qrels,
            se=self,
            metric=metric,
            n_trials=n_trials,
            cutoff=cutoff,
        )
        self.hyperparams = hyperparams
        self.save()



import os
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Set, Union

import numba as nb
import numpy as np
import orjson
from numba.typed import List as TypedList
from oneliner_utils import create_path, read_jsonl
from tqdm import tqdm

from ..autotune import tune_bm25
from ..base_retriever import BaseRetriever
from ..paths import docs_path, fr_state_path
from ..sparse_retriever.preprocessing import (
    get_stemmer,
    get_stopwords,
    get_tokenizer,
    preprocessing,
    preprocessing_multi,
)
from ..sparse_retriever.sparse_retrieval_models.bm25 import bm25, bm25_multi
from ..sparse_retriever.sparse_retrieval_models.tf_idf import tf_idf, tf_idf_multi
from ..sparse_retriever.sparse_retriever import SparseRetriever, build_inverted_index
from ..utils.numba_utils import diff_sorted, intersect_sorted_multi, union_sorted_multi

CLAUSE_LIST = ["must", "must not"]
OPERATOR_LIST = ["eq", "gt", "gte", "lt", "lte", "between", "and", "or"]
KIND_LIST = ["id", "text", "number", "bool", "keyword", "keywords"]


class AdvancedRetriever(SparseRetriever):
    def __init__(
        self,
        schema: Dict[str, str],
        index_name: str = "new-index",
        model: str = "bm25",
        min_df: int = 1,
        tokenizer: Union[str, callable] = "whitespace",
        stemmer: Union[str, callable] = "english",
        stopwords: Union[str, List[str], Set[str]] = "english",
        do_lowercasing: bool = True,
        do_ampersand_normalization: bool = True,
        do_special_chars_normalization: bool = True,
        do_acronyms_normalization: bool = True,
        do_punctuation_removal: bool = True,
        hyperparams: dict = None,
    ):
        assert model.lower() in {"bm25", "tf-idf"}
        assert min_df > 0, "`min_df` must be greater than zero."
        self.check_schema(schema)
        self.init_args = {
            "schema": schema,
            "model": model.lower(),
            "min_df": min_df,
            "index_name": index_name,
            "do_lowercasing": do_lowercasing,
            "do_ampersand_normalization": do_ampersand_normalization,
            "do_special_chars_normalization": do_special_chars_normalization,
            "do_acronyms_normalization": do_acronyms_normalization,
            "do_punctuation_removal": do_punctuation_removal,
            "tokenizer": tokenizer,
            "stemmer": stemmer,
            "stopwords": stopwords,
        }

        self.schema = schema
        self.text_field = [k for k, v in self.schema.items() if v == "text"][0]
        self.model = model.lower()
        self.min_df = min_df
        self.index_name = index_name
        self.do_lowercasing = do_lowercasing
        self.do_ampersand_normalization = do_ampersand_normalization
        self.do_special_chars_normalization = do_special_chars_normalization
        self.do_acronyms_normalization = do_acronyms_normalization
        self.do_punctuation_removal = do_punctuation_removal
        self.tokenizer = get_tokenizer(tokenizer)
        self.stemmer = get_stemmer(stemmer)
        self.stopwords = [self.stemmer(sw) for sw in get_stopwords(stopwords)]
        self.id_mapping = None
        self.reversed_id_mapping = None
        self.inverted_index = None
        self.vocabulary = None
        self.doc_count = None
        self.doc_ids = None
        self.doc_lens = None
        self.avg_doc_len = None
        self.relative_doc_lens = None
        self.doc_index = None
        # self.mappings = None
        self.metadata = None

        self.preprocessing_kwargs = {
            "tokenizer": self.tokenizer,
            "stemmer": self.stemmer,
            "stopwords": self.stopwords,
            "do_lowercasing": self.do_lowercasing,
            "do_ampersand_normalization": self.do_ampersand_normalization,
            "do_special_chars_normalization": self.do_special_chars_normalization,
            "do_acronyms_normalization": self.do_acronyms_normalization,
            "do_punctuation_removal": self.do_punctuation_removal,
        }

        self.preprocessing_pipe = preprocessing_multi(**self.preprocessing_kwargs)

        self.hyperparams = dict(b=0.75, k1=1.2) if hyperparams is None else hyperparams

    def save(self) -> None:
        """Save the state of the retriever to be able to restore it later."""

        state = {
            "init_args": self.init_args,
            "id_mapping": self.id_mapping,
            "doc_count": self.doc_count,
            "inverted_index": self.inverted_index,
            "vocabulary": self.vocabulary,
            "doc_lens": self.doc_lens,
            "relative_doc_lens": self.relative_doc_lens,
            "hyperparams": self.hyperparams,
        }

        np.savez_compressed(fr_state_path(self.index_name), state=state)

    @staticmethod
    def load(index_name: str = "new-index"):
        """Load a retriever and its index.

        Args:
            index_name (str, optional): Name of the index. Defaults to "new-index".

        Returns:
            SparseRetriever: Sparse Retriever.
        """

        state = np.load(fr_state_path(index_name), allow_pickle=True)["state"][()]

        se = AdvancedRetriever(**state["init_args"])
        se.initialize_doc_index()
        se.id_mapping = state["id_mapping"]
        se.reversed_id_mapping = {v: k for k, v in state["id_mapping"].items()}
        se.doc_count = state["doc_count"]
        se.doc_ids = np.arange(state["doc_count"], dtype=np.int32)
        se.inverted_index = state["inverted_index"]
        se.vocabulary = set(se.inverted_index)
        se.doc_lens = state["doc_lens"]
        se.relative_doc_lens = state["relative_doc_lens"]
        se.hyperparams = state["hyperparams"]

        state = {
            "init_args": se.init_args,
            "id_mapping": se.id_mapping,
            "doc_count": se.doc_count,
            "inverted_index": se.inverted_index,
            "vocabulary": se.vocabulary,
            "doc_lens": se.doc_lens,
            "relative_doc_lens": se.relative_doc_lens,
            "hyperparams": se.hyperparams,
        }

        return se

    def check_schema(self, schema: Dict[str, str]) -> None:
        """Check if schema is valid"""
        text_found = False

        if "id" not in schema:
            raise ValueError("Schema must contain an id field")

        for k in schema:
            if not isinstance(k, str):
                raise TypeError("Schema keys must be strings")

        for value in schema.values():
            if value not in KIND_LIST:
                raise ValueError(f"Type {value} not supported")
            if value == "text":
                if text_found:
                    raise ValueError("Only one field can be text")
                text_found = True

        return True

    def check_collection(self, collection: Iterable, schema: Dict[str, str]) -> None:
        """Check collection against a schema"""
        for i, doc in enumerate(collection):
            if "id" not in doc:
                raise ValueError(f"Doc #{i} has no id")

            doc_id = doc["id"]

            for field in schema:
                if field not in doc:
                    raise ValueError(f"Field {field} not in doc {doc_id}")

            for field in doc:
                if field not in schema:
                    raise ValueError(f"Field {field} not in schema")

                kind = schema[field]
                value = doc[field]

                if kind == "id" and not isinstance(value, (int, str)):
                    raise TypeError(f"Field {field} of doc #{i} has wrong type")

                elif kind == "text" and not isinstance(value, str):
                    raise TypeError(f"Field {field} of doc {doc_id} has wrong type")

                elif kind == "number" and not isinstance(value, (int, float)):
                    raise TypeError(f"Field {field} of doc {doc_id} has wrong type")

                elif kind == "bool" and not isinstance(value, bool):
                    raise TypeError(f"Field {field} of doc {doc_id} has wrong type")

                elif kind == "keyword" and not isinstance(value, str):
                    raise TypeError(f"Field {field} of doc {doc_id} has wrong type")

                elif kind == "keywords" and not isinstance(value, (list, set, tuple)):
                    raise TypeError(f"Field {field} of doc {doc_id} has wrong type")

        return True

    def initialize_metadata(self, schema):
        metadata = {}

        for field, kind in schema.items():
            if kind == "number":
                metadata[field] = []
            if kind == "bool":
                metadata[field] = {True: [], False: []}
            elif kind in {"keyword", "keywords"}:
                metadata[field] = defaultdict(list)

        return metadata

    def fill_metadata(self, metadata, collection, schema):
        for i, doc in enumerate(collection):
            for field, kind in schema.items():
                if kind == "number":
                    metadata[field].append(doc[field])
                elif kind in ["bool", "keyword"]:
                    metadata[field][doc[field]].append(i)
                elif kind == "keywords":
                    for keyword in doc[field]:
                        metadata[field][keyword].append(i)

        return metadata

    def index_metadata(self, collection, schema):
        metadata = self.initialize_metadata(schema)
        metadata = self.fill_metadata(metadata, collection, schema)

        # Convert to numpy arrays
        for field, kind in schema.items():
            if kind == "number":
                metadata[field] = np.array(metadata[field])
            elif kind == "bool":
                metadata[field][True] = np.array(metadata[field][True], dtype=np.int32)
                metadata[field][False] = np.array(
                    metadata[field][False], dtype=np.int32
                )
            elif kind in ["keyword", "keywords"]:
                metadata[field] = dict(metadata[field])
                for keyword in metadata[field]:
                    metadata[field][keyword] = np.array(
                        metadata[field][keyword], dtype=np.int32
                    )
        return metadata

    def index_aux(self, text_field: str, show_progress: bool = True):
        """Internal usage."""
        collection = read_jsonl(
            docs_path(self.index_name),
            generator=True,
            callback=lambda x: x[text_field],
        )

        # Preprocessing --------------------------------------------------------
        collection = self.preprocessing_pipe(collection, generator=True)

        # Inverted index -------------------------------------------------------
        (
            self.inverted_index,
            self.doc_lens,
            self.relative_doc_lens,
        ) = build_inverted_index(
            collection=collection,
            n_docs=self.doc_count,
            min_df=self.min_df,
            show_progress=show_progress,
        )
        self.avg_doc_len = np.mean(self.doc_lens, dtype=np.float32)
        self.vocabulary = set(self.inverted_index)

    def index(
        self,
        collection: Iterable,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index a given collection of documents.

        Args:
            collection (Iterable): collection of documents to index.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            SparseRetriever: Sparse Retriever.
        """
        self.check_collection(collection, self.schema)
        self.save_collection(collection, callback)
        self.initialize_doc_index()
        self.initialize_id_mapping()
        self.reversed_id_mapping = {v: k for k, v in self.id_mapping.items()}
        self.doc_count = len(self.id_mapping)
        self.doc_ids = np.arange(self.doc_count, dtype=np.int32)
        self.index_aux(
            text_field=self.text_field,
            show_progress=show_progress,
        )
        self.metadata = self.index_metadata(collection, schema=self.schema)
        self.save()
        return self

    def index_file(
        self, path: str, callback: callable = None, show_progress: bool = True
    ):
        """Index the collection contained in a given file.

        Args:
            path (str): path of file containing the collection to index.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            SparseRetriever: Sparse Retriever
        """

        # collection = self.collection_generator(path=path, callback=callback)
        collection_generator = self.collection_generator

        class Collection:
            def __init__(self, path, callback):
                self.path = path
                self.callback = callback

            def __iter__(self):
                yield from collection_generator(path=self.path, callback=self.callback)

        return self.index(
            collection=Collection(path, callback), show_progress=show_progress
        )

    def filter_doc_ids(
        self,
        field: str,
        clause: str,
        value: Any = None,
        operator: str = None,
        raise_error: bool = True,
    ):
        if clause not in CLAUSE_LIST:
            raise ValueError(f"Clause must be one of {CLAUSE_LIST}")
        if operator is not None and operator not in OPERATOR_LIST:
            raise ValueError(f"Operator must be one of {OPERATOR_LIST}")
        if field not in self.schema:
            raise ValueError(f"Field `{field}` not in schema")

        kind = self.schema[field]
        doc_ids = self.doc_ids
        id_mapping = self.reversed_id_mapping
        metadata = self.metadata

        def get_value(field, value):
            if raise_error and value not in metadata[field]:
                raise ValueError(f"No document has value `{value}` in field `{field}`.")

            return metadata[field].get(value, np.array([], dtype=np.int32))

        if kind == "id":
            if clause == "must":
                return doc_ids[np.isin(doc_ids, [id_mapping[v] for v in value])]
            elif clause == "must not":
                return doc_ids[~np.isin(doc_ids, [id_mapping[v] for v in value])]

        elif kind == "bool":
            if clause == "must":
                return metadata[field][value]
            elif clause == "must not":
                return metadata[field][not value]

        elif kind == "keyword":
            if clause == "must":
                if isinstance(value, list):
                    return union_sorted_multi(
                        TypedList([get_value(field, v) for v in value])
                    )
                else:
                    return get_value(field, value)

            elif clause == "must not":
                if isinstance(value, list):
                    ids = [
                        get_value(field, v) for v in metadata[field] if v not in value
                    ]

                else:
                    ids = [get_value(field, v) for v in metadata[field] if v != value]

                return union_sorted_multi(TypedList(ids))

        elif kind == "keywords":
            if clause == "must":
                if isinstance(value, list):
                    if operator == "and":
                        return intersect_sorted_multi(
                            TypedList([get_value(field, v) for v in value])
                        )
                    elif operator == "or":
                        return union_sorted_multi(
                            TypedList([get_value(field, v) for v in value])
                        )
                    else:
                        raise ValueError(
                            f"Operator `{operator}`not supported for keywords field"
                        )
                else:
                    return get_value(field, value)

            elif clause == "must not":
                if isinstance(value, list):
                    must_not_ids = [get_value(field, v) for v in value]
                    if operator == "and":
                        must_not_ids = intersect_sorted_multi(TypedList(must_not_ids))
                    elif operator == "or":
                        must_not_ids = union_sorted_multi(TypedList(must_not_ids))
                    else:
                        raise ValueError(
                            f"Operator `{operator}`not supported for keywords field"
                        )

                    return diff_sorted(
                        np.arange(self.doc_count, dtype=np.int32), must_not_ids
                    )

                else:
                    return diff_sorted(
                        np.arange(self.doc_count, dtype=np.int32),
                        get_value(field, value),
                    )

        elif kind == "number":
            if operator == "eq":
                mask = metadata[field] == value
            elif operator == "gt":
                mask = metadata[field] > value
            elif operator == "gte":
                mask = metadata[field] >= value
            elif operator == "lt":
                mask = metadata[field] < value
            elif operator == "lte":
                mask = metadata[field] <= value
            elif operator == "between":
                data, min_v, max_v = metadata[field], value[0], value[1]
                mask = np.logical_and(data >= min_v, data <= max_v)
            else:
                raise ValueError("Operator not supported for numeric field")

            if clause == "must":
                return doc_ids[mask]
            elif clause == "must not":
                return doc_ids[~mask]

        else:
            raise ValueError(
                f"Field {field} of type {kind} not supported for filtering"
            )

    def get_filtered_doc_ids(self, filters: List[Dict]) -> np.ndarray:
        if len(filters) == 1:
            return self.filter_doc_ids(**filters[0])
        filtered_doc_ids = TypedList([self.filter_doc_ids(**f) for f in filters])
        return intersect_sorted_multi(filtered_doc_ids)

    def format_filters(self, filters: Dict, clause: str = "must") -> List[Dict]:
        formatted_filters = []

        for field, value in filters.items():
            if self.schema[field] in {"id", "bool", "keyword"}:
                f = dict(field=field, clause=clause, value=value)

            elif self.schema[field] in {"number", "keywords"}:
                f = dict(field=field, clause=clause, value=value[1], operator=value[0])

            formatted_filters.append(f)

        return formatted_filters

    def search(
        self,
        query: Union[Dict, str],
        return_docs: bool = True,
        cutoff: int = 100,
        operator: str = "OR",
        subset_doc_ids: List = None,
    ) -> List:
        if isinstance(query, str):
            query_text = query
            if subset_doc_ids is not None:
                subset_doc_ids = np.array(
                    [self.reversed_id_mapping[doc_id] for doc_id in subset_doc_ids],
                    dtype=np.int32,
                )
        else:
            query_text = query.get("text", "")
            must_filters = query.get("where", {})
            must_not_filters = query.get("where_not", {})
            must_single_filters = {
                k: v
                for k, v in query.items()
                if k not in {"text", "where", "where_not"}
            }

            must_filters = self.format_filters(must_filters)
            must_not_filters = self.format_filters(must_not_filters, clause="must not")
            must_single_filters = self.format_filters(must_single_filters)
            filters = must_filters + must_not_filters + must_single_filters
            subset_doc_ids = self.get_filtered_doc_ids(filters)

        query_terms = self.query_preprocessing(query_text)
        query_terms = [t for t in query_terms if t in self.vocabulary]

        if query_terms:
            doc_ids = self.get_doc_ids(query_terms)
            term_doc_freqs = self.get_term_doc_freqs(query_terms)

            if self.model == "bm25":
                unique_doc_ids, scores = bm25(
                    term_doc_freqs=term_doc_freqs,
                    doc_ids=doc_ids,
                    relative_doc_lens=self.relative_doc_lens,
                    doc_count=self.doc_count,
                    cutoff=cutoff,
                    operator=operator,
                    subset_doc_ids=subset_doc_ids,
                    **self.hyperparams,
                )
            elif self.model == "tf-idf":
                unique_doc_ids, scores = tf_idf(
                    term_doc_freqs=term_doc_freqs,
                    doc_ids=doc_ids,
                    doc_lens=self.doc_lens,
                    cutoff=cutoff,
                    operator=operator,
                    subset_doc_ids=subset_doc_ids,
                )
            else:
                raise NotImplementedError()
        else:
            if subset_doc_ids is None:
                unique_doc_ids = self.doc_ids
                scores = np.ones(self.doc_count)
            else:
                unique_doc_ids = subset_doc_ids
                scores = np.ones(len(subset_doc_ids))

        unique_doc_ids = self.map_internal_ids_to_original_ids(unique_doc_ids)

        if not return_docs:
            return dict(zip(unique_doc_ids, scores))

        return self.prepare_results(unique_doc_ids, scores)



from typing import Dict, List


def extract_scores(results):
    """Extract the scores from a given results dictionary."""
    scores = [None] * len(results)
    for i, v in enumerate(results.values()):
        scores[i] = v
    return scores


def safe_max(x):
    return max(x) if len(x) != 0 else 0


def safe_min(x):
    return min(x) if len(x) != 0 else 0


def min_max_norm(run: Dict[str, float]):
    """Apply `min-max normalization` to a given run."""
    normalized_run = {}

    for q_id, results in run.items():
        scores = extract_scores(results)
        min_score = safe_min(scores)
        max_score = safe_max(scores)
        denominator = max(max_score - min_score, 1e-9)

        normalized_results = {
            doc_id: (results[doc_id] - min_score) / (denominator) for doc_id in results
        }

        normalized_run[q_id] = normalized_results

    return normalized_run


def max_norm(run: Dict[str, float]):
    """Apply `max normalization` to a given run."""
    normalized_run = {}

    for q_id, results in run.items():
        scores = extract_scores(results)
        max_score = safe_max(scores)
        denominator = max(max_score, 1e-9)

        normalized_results = {
            doc_id: results[doc_id] / denominator for doc_id in results
        }

        normalized_run[q_id] = normalized_results

    return normalized_run


def sum_norm(run: Dict[str, float]):
    """Apply `sum normalization` to a given run."""
    normalized_run = {}

    for q_id, results in run.items():
        scores = extract_scores(results)
        min_score = safe_min(scores)
        sum_score = sum(scores)
        denominator = sum_score - min_score * len(results)
        denominator = max(denominator, 1e-9)

        normalized_results = {
            doc_id: (results[doc_id] - min_score) / (denominator) for doc_id in results
        }

        normalized_run[q_id] = normalized_results

    return normalized_run


def min_max_norm_multi(runs: List[Dict[str, float]]):
    """Apply `min-max normalization` to a list of given runs."""
    return [min_max_norm(run) for run in runs]


def max_norm_multi(runs: List[Dict[str, float]]):
    """Apply `max normalization` to a list of given runs."""
    return [max_norm(run) for run in runs]


def sum_norm_multi(runs: List[Dict[str, float]]):
    """Apply `sum normalization` to a list of given runs."""
    return [sum_norm(run) for run in runs]



from collections import defaultdict
from typing import Dict, List

import numpy as np

from ..autotune import tune_merger
from ..paths import merger_state_path
from .normalization import max_norm_multi, min_max_norm_multi, sum_norm_multi


class Merger:
    def __init__(self, index_name: str = "new-index"):
        self.index_name = index_name
        self.norm = "min-max"
        self.params = None

    def fuse(
        self, results: List[Dict[str, float]], cutoff: int = 100
    ) -> Dict[str, float]:
        return self.mfuse([{"q_0": res} for res in results], cutoff)["q_0"]

    def mfuse(
        self, runs: List[Dict[str, Dict[str, float]]], cutoff: int = 100
    ) -> Dict[str, Dict[str, float]]:
        if self.norm == "min-max":
            normalized_runs = min_max_norm_multi(runs)
        elif self.norm == "max":
            normalized_runs = max_norm_multi(runs)
        elif self.norm == "sum":
            normalized_runs = sum_norm_multi(runs)
        else:
            raise NotImplementedError

        weights = [1.0 for _ in runs] if self.params is None else self.params["weights"]

        fused_run = defaultdict(lambda: defaultdict(float))
        for i, run in enumerate(normalized_runs):
            for q_id in run:
                for doc_id in run[q_id]:
                    fused_run[q_id][doc_id] += weights[i] * run[q_id][doc_id]

        # Sort results by descending value and ascending key
        for q_id, results in list(fused_run.items()):
            fused_run[q_id] = dict(sorted(results.items(), key=lambda x: (-x[1], x[0])))

        # Apply cutoff
        for q_id, results in list(fused_run.items()):
            fused_run[q_id] = dict(list(results.items())[:cutoff])

        return dict(fused_run)

    def save(self):
        state = dict(
            init_args=dict(index_name=self.index_name),
            norm=self.norm,
            params=self.params,
        )
        np.savez_compressed(merger_state_path(self.index_name), state=state)

    @staticmethod
    def load(index_name: str = "new-index"):
        state = np.load(merger_state_path(index_name), allow_pickle=True)["state"][()]
        merger = Merger(**state["init_args"])
        merger.norm = state["norm"]
        merger.params = state["params"]
        return merger

    def autotune(
        self,
        qrels: Dict[str, Dict[str, float]],
        runs: List[Dict[str, Dict[str, float]]],
        metric: str = "ndcg",
    ):
        config = tune_merger(qrels, runs, metric)
        self.norm = config["norm"]
        self.params = config["params"]
        self.save()



import os
import shutil
from typing import Iterable, List

import numpy as np
import orjson
from indxr import Indxr
from oneliner_utils import read_csv, read_jsonl

from .paths import docs_path, index_path


class BaseRetriever:
    def __init__(self, index_name: str = "new-index"):
        self.index_name = index_name
        self.id_mapping = None
        self.doc_count = None
        self.doc_index = None

    @staticmethod
    def delete(index_name="new-index"):
        try:
            shutil.rmtree(index_path(index_name))
            print(f"{index_name} successfully removed.")
        except FileNotFoundError:
            print(f"{index_name} not found.")

    def collection_generator(self, path: str, callback: callable = None):
        kind = os.path.splitext(path)[1][1:]
        assert kind in {
            "jsonl",
            "csv",
            "tsv",
        }, "Only JSONl, CSV, and TSV are currently supported."

        if kind == "jsonl":
            collection = read_jsonl(path, generator=True, callback=callback)
        elif kind == "csv":
            collection = read_csv(path, generator=True, callback=callback)
        elif kind == "tsv":
            collection = read_csv(
                path, delimiter="\t", generator=True, callback=callback
            )

        return collection

    def save_collection(self, collection: Iterable, callback: callable = None):
        with open(docs_path(self.index_name), "wb") as f:
            for doc in collection:
                x = callback(doc) if callback is not None else doc
                f.write(orjson.dumps(x) + "\n".encode())

    def initialize_doc_index(self):
        self.doc_index = Indxr(docs_path(self.index_name))

    def initialize_id_mapping(self):
        ids = read_jsonl(
            docs_path(self.index_name),
            generator=True,
            callback=lambda x: x["id"],
        )
        self.id_mapping = dict(enumerate(ids))

    def get_doc(self, doc_id: str) -> dict:
        return self.doc_index.get(doc_id)

    def get_docs(self, doc_ids: List[str]) -> List[dict]:
        return self.doc_index.mget(doc_ids)

    def prepare_results(self, doc_ids: List[str], scores: np.ndarray) -> List[dict]:
        docs = self.get_docs(doc_ids)
        results = []
        for doc, score in zip(docs, scores):
            doc["score"] = score
            results.append(doc)

        return results

    def map_internal_ids_to_original_ids(self, doc_ids: Iterable) -> List[str]:
        return [self.id_mapping[doc_id] for doc_id in doc_ids]

    def save(self):
        raise NotImplementedError()

    @staticmethod
    def load():
        raise NotImplementedError()

    def index(self):
        raise NotImplementedError()

    def index_file(self):
        raise NotImplementedError()

    def search(self):
        raise NotImplementedError()

    def msearch(self):
        raise NotImplementedError()

    def autotune(self):
        raise NotImplementedError()



__all__ = [
    "ANN_Searcher",
    "DenseRetriever",
    "Encoder",
    "SearchEngine",
    "SparseRetriever",
    "HybridRetriever",
    "Merger",
]

import os
from pathlib import Path

from .dense_retriever.ann_searcher import ANN_Searcher
from .dense_retriever.dense_retriever import DenseRetriever
from .dense_retriever.encoder import Encoder
from .hybrid_retriever import HybridRetriever
from .merger.merger import Merger
from .sparse_retriever.sparse_retriever import SparseRetriever
from .sparse_retriever.sparse_retriever import SparseRetriever as SearchEngine

# Set environment variables ----------------------------------------------------
os.environ["TOKENIZERS_PARALLELISM"] = "false"
os.environ["RETRIV_BASE_PATH"] = str(Path.home() / ".retriv")


def set_base_path(path: str):
    os.environ["RETRIV_BASE_PATH"] = path



from typing import Dict, Iterable, List, Set, Union

import numpy as np
import orjson
from oneliner_utils import create_path
from tqdm import tqdm

from .base_retriever import BaseRetriever
from .dense_retriever.dense_retriever import DenseRetriever
from .merger.merger import Merger
from .paths import hr_state_path
from .sparse_retriever.sparse_retriever import SparseRetriever


class HybridRetriever(BaseRetriever):
    def __init__(
        self,
        # Global params
        index_name: str = "new-index",
        # Sparse retriever params
        sr_model: str = "bm25",
        min_df: int = 1,
        tokenizer: Union[str, callable] = "whitespace",
        stemmer: Union[str, callable] = "english",
        stopwords: Union[str, List[str], Set[str]] = "english",
        do_lowercasing: bool = True,
        do_ampersand_normalization: bool = True,
        do_special_chars_normalization: bool = True,
        do_acronyms_normalization: bool = True,
        do_punctuation_removal: bool = True,
        # Dense retriever params
        dr_model: str = "sentence-transformers/all-MiniLM-L6-v2",
        normalize: bool = True,
        max_length: int = 128,
        use_ann: bool = True,
        # For already instantiated modules
        sparse_retriever: SparseRetriever = None,
        dense_retriever: DenseRetriever = None,
        merger: Merger = None,
    ):
        """The [Hybrid Retriever](https://github.com/AmenRa/retriv/blob/main/docs/hybrid_retriever.md) is searcher based on both lexical and semantic matching. It comprises three components: the [Sparse Retriever]((https://github.com/AmenRa/retriv/blob/main/docs/sparse_retriever.md)), the [Dense Retriever]((https://github.com/AmenRa/retriv/blob/main/docs/dense_retriever.md)), and the Merger. The Merger fuses the results of the Sparse and Dense Retrievers to compute the _hybrid_ results.

        Args:
            index_name (str, optional): [retriv](https://github.com/AmenRa/retriv) will use `index_name` as the identifier of your index. Defaults to "new-index".

            sr_model (str, optional): defines the model to use for sparse retrieval (`bm25` or `tf-idf`). Defaults to "bm25".

            min_df (int, optional): terms that appear in less than `min_df` documents will be ignored. If integer, the parameter indicates the absolute count. If float, it represents a proportion of documents. Defaults to 1.

            tokenizer (Union[str, callable], optional): [tokenizer](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to use during preprocessing. You can pass a custom callable tokenizer or disable tokenization setting the parameter to `None`. Defaults to "whitespace".

            stemmer (Union[str, callable], optional): [stemmer](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to use during preprocessing. You can pass a custom callable stemmer or disable stemming setting the parameter to `None`. Defaults to "english".

            stopwords (Union[str, List[str], Set[str]], optional): [stopwords](https://github.com/AmenRa/retriv/blob/main/docs/text_preprocessing.md) to remove during preprocessing. You can pass a custom stop-word list or disable stop-words removal by setting the parameter to `None`. Defaults to "english".

            do_lowercasing (bool, optional): whether to lowercase texts. Defaults to True.

            do_ampersand_normalization (bool, optional): whether to convert `&` in `and` during pre-processing. Defaults to True.

            do_special_chars_normalization (bool, optional): whether to remove special characters for letters, e.g., `übermensch` → `ubermensch`. Defaults to True.

            do_acronyms_normalization (bool, optional): whether to remove full stop symbols from acronyms without splitting them in multiple words, e.g., `P.C.I.` → `PCI`. Defaults to True.

            do_punctuation_removal (bool, optional): whether to remove punctuation. Defaults to True.

            dr_model (str, optional): defines the model to use for encoding queries and documents into vectors. You can use an [HuggingFace's Transformers](https://huggingface.co/models) pre-trained model by providing its ID or load a local model by providing its path. In the case of local models, the path must point to the directory containing the data saved with the [`PreTrainedModel.save_pretrained`](https://huggingface.co/docs/transformers/v4.26.1/en/main_classes/model#transformers.PreTrainedModel.save_pretrained) method. Note that the representations are computed with `mean pooling` over the `last_hidden_state`. Defaults to "sentence-transformers/all-MiniLM-L6-v2".

            normalize (bool, optional):  whether to L2 normalize the vector representations. Defaults to True.

            max_length (int, optional): texts longer than `max_length` will be automatically truncated. Choose this parameter based on how the employed model was trained or is generally used. Defaults to 128.

            use_ann (bool, optional): whether to use approximate nearest neighbors search. Set it to `False` to use nearest neighbors search without approximation. If you have less than 20k documents in your collection, you probably want to disable approximation. Defaults to True.
        """

        self.index_name = index_name

        self.sparse_retriever = (
            sparse_retriever
            if sparse_retriever is not None
            else SparseRetriever(
                index_name=index_name,
                model=sr_model,
                min_df=min_df,
                tokenizer=tokenizer,
                stemmer=stemmer,
                stopwords=stopwords,
                do_lowercasing=do_lowercasing,
                do_ampersand_normalization=do_ampersand_normalization,
                do_special_chars_normalization=do_special_chars_normalization,
                do_acronyms_normalization=do_acronyms_normalization,
                do_punctuation_removal=do_punctuation_removal,
            )
        )

        self.dense_retriever = (
            dense_retriever
            if dense_retriever is not None
            else DenseRetriever(
                index_name=index_name,
                model=dr_model,
                normalize=normalize,
                max_length=max_length,
                use_ann=use_ann,
            )
        )

        self.merger = merger if merger is not None else Merger(index_name=index_name)

    def index(
        self,
        collection: Iterable,
        embeddings_path: str = None,
        use_gpu: bool = False,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index a given collection of documents.

        Args:
            collection (Iterable): collection of documents to index.

            embeddings_path (str, optional): in case you want to load pre-computed embeddings, you can provide the path to a `.npy` file. Embeddings must be in the same order as the documents in the collection file. Defaults to None.

            use_gpu (bool, optional): whether to use the GPU for document encoding. Defaults to False.

            batch_size (int, optional): how many documents to encode at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 512.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            HybridRetriever: Hybrid Retriever
        """

        self.save_collection(collection, callback)

        self.initialize_doc_index()
        self.initialize_id_mapping()
        self.doc_count = len(self.id_mapping)

        # Sparse ---------------------------------------------------------------
        self.sparse_retriever.doc_index = self.doc_index
        self.sparse_retriever.id_mapping = self.id_mapping
        self.sparse_retriever.doc_count = self.doc_count
        self.sparse_retriever.index_aux(show_progress)

        # Dense ----------------------------------------------------------------
        self.dense_retriever.doc_index = self.doc_index
        self.dense_retriever.id_mapping = self.id_mapping
        self.dense_retriever.doc_count = self.doc_count
        self.dense_retriever.index_aux(
            embeddings_path, use_gpu, batch_size, callback, show_progress
        )

        self.save()

        return self

    def index_file(
        self,
        path: str,
        embeddings_path: str = None,
        use_gpu: bool = False,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index the collection contained in a given file.

        Args:
            path (str): path of file containing the collection to index.

            embeddings_path (str, optional): in case you want to load pre-computed embeddings, you can provide the path to a `.npy` file. Embeddings must be in the same order as the documents in the collection file. Defaults to None.

            use_gpu (bool, optional): whether to use the GPU for document encoding. Defaults to False.

            batch_size (int, optional): how many documents to encode at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 512.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            HybridRetriever: Hybrid Retriever.
        """

        collection = self.collection_generator(path, callback)
        return self.index(
            collection,
            embeddings_path,
            use_gpu,
            batch_size,
            None,
            show_progress,
        )

    def save(self):
        """Save the state of the retriever to be able to restore it later."""

        state = dict(
            id_mapping=self.id_mapping,
            doc_count=self.doc_count,
        )
        np.savez_compressed(hr_state_path(self.index_name), state=state)

        self.sparse_retriever.save()
        self.dense_retriever.save()
        self.merger.save()

    @staticmethod
    def load(index_name: str = "new-index"):
        """Load a retriever and its index.

        Args:
            index_name (str, optional): Name of the index. Defaults to "new-index".

        Returns:
            HybridRetriever: Hybrid Retriever.
        """

        state = np.load(hr_state_path(index_name), allow_pickle=True)["state"][()]

        hr = HybridRetriever(index_name)
        hr.initialize_doc_index()
        hr.id_mapping = state["id_mapping"]
        hr.doc_count = state["doc_count"]

        hr.sparse_retriever = SparseRetriever.load(index_name)
        hr.dense_retriever = DenseRetriever.load(index_name)
        hr.merger = Merger.load(index_name)
        return hr

    def search(
        self,
        query: str,
        return_docs: bool = True,
        cutoff: int = 100,
    ) -> List:
        """Standard search functionality.

        Args:
            query (str): what to search for.

            return_docs (bool, optional): whether to return the texts of the documents. Defaults to True.

            cutoff (int, optional): number of results to return. Defaults to 100.

        Returns:
            List: results.
        """

        sparse_results = self.sparse_retriever.search(query, False, 1_000)
        dense_results = self.dense_retriever.search(query, False, 1_000)
        hybrid_results = self.merger.fuse([sparse_results, dense_results])
        return (
            self.prepare_results(
                list(hybrid_results.keys())[:cutoff],
                list(hybrid_results.values())[:cutoff],
            )
            if return_docs
            else hybrid_results
        )

    def msearch(
        self,
        queries: List[Dict[str, str]],
        cutoff: int = 100,
        batch_size: int = 32,
    ) -> Dict:
        """Compute results for multiple queries at once.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

            batch_size (int, optional): how many queries to search at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 32.

        Returns:
            Dict: results.
        """

        sparse_results = self.sparse_retriever.msearch(queries, 1_000)
        dense_results = self.dense_retriever.msearch(queries, 1_000, batch_size)
        return self.merger.mfuse([sparse_results, dense_results], cutoff)

    def bsearch(
        self,
        queries: List[Dict[str, str]],
        cutoff: int = 100,
        batch_size: int = 32,
        show_progress: bool = True,
        qrels: Dict[str, Dict[str, float]] = None,
        path: str = None,
    ):
        """Batch-Search is similar to Multi-Search but automatically generates batches of queries to evaluate and allows dynamic writing of the search results to disk in [JSONl](https://jsonlines.org) format. bsearch is handy for computing results for hundreds of thousands or even millions of queries without hogging your RAM.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

            batch_size (int, optional): how many queries to search at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 32.

            show_progress (bool, optional): whether to show a progress bar for the search process. Defaults to True.

            qrels (Dict[str, Dict[str, float]], optional): query relevance judgements for the queries. Defaults to None.

            path (str, optional): where to save the results. Defaults to None.

        Returns:
            Dict: results.
        """

        batches = [
            queries[i : i + batch_size] for i in range(0, len(queries), batch_size)
        ]

        results = {}

        pbar = tqdm(
            total=len(queries),
            disable=not show_progress,
            desc="Batch search",
            dynamic_ncols=True,
            mininterval=0.5,
        )

        if path is None:
            for batch in batches:
                new_results = self.msearch(
                    queries=batch, cutoff=cutoff, batch_size=len(batch)
                )
                results = {**results, **new_results}
                pbar.update(min(batch_size, len(batch)))
        else:
            path = create_path(path)
            path.parent.mkdir(parents=True, exist_ok=True)

            with open(path, "wb") as f:
                for batch in batches:
                    new_results = self.msearch(queries=batch, cutoff=cutoff)

                    for i, (k, v) in enumerate(new_results.items()):
                        x = {
                            "id": k,
                            "text": batch[i]["text"],
                            "hybrid_doc_ids": list(v.keys()),
                            "hybrid_scores": [float(s) for s in list(v.values())],
                        }
                        if qrels is not None:
                            x["rel_doc_ids"] = list(qrels[k].keys())
                            x["rel_scores"] = list(qrels[k].values())
                        f.write(orjson.dumps(x) + "\n".encode())

                    pbar.update(min(batch_size, len(batch)))

        return results

    def autotune(
        self,
        queries: List[Dict[str, str]],
        qrels: Dict[str, Dict[str, float]],
        metric: str = "ndcg",
        n_trials: int = 100,
        cutoff: int = 100,
        batch_size: int = 32,
    ):
        """Use the AutoTune function to tune the Sparse Retriever's model [BM25](https://en.wikipedia.org/wiki/Okapi_BM25) parameters and the importance given to the lexical and semantic relevance scores computed by the Sparse and Dense Retrievers, respectively. All metrics supported by [ranx](https://github.com/AmenRa/ranx) are supported by the `autotune` function. At the of the process, the best parameter configurations are automatically applied and saved to disk. You can inspect the best configurations found by printing `hr.sparse_retriever.hyperparams`, `hr.merger.norm` and `hr.merger.params`.

        Args:
            queries (List[Dict[str, str]]): queries to use for the optimization process.

            qrels (Dict[str, Dict[str, float]]): query relevance judgements for the queries.

            metric (str, optional): metric to optimize for. Defaults to "ndcg".

            n_trials (int, optional): number of configuration to evaluate. Defaults to 100.

            cutoff (int, optional): number of results to consider for the optimization process. Defaults to 100.
        """

        # Tune sparse ----------------------------------------------------------
        self.sparse_retriever.autotune(
            queries=queries,
            qrels=qrels,
            metric=metric,
            n_trials=n_trials,
            cutoff=cutoff,
        )

        # Tune merger ----------------------------------------------------------
        sparse_results = self.sparse_retriever.msearch(queries, 1_000)
        dense_results = self.dense_retriever.msearch(queries, 1_000, batch_size)
        self.merger.autotune(qrels, [sparse_results, dense_results], metric)



import os
import shutil
from typing import Dict, Iterable, List

import numpy as np
import orjson
from numba import njit, prange
from numba.typed import List as TypedList
from oneliner_utils import create_path
from tqdm import tqdm

from ..base_retriever import BaseRetriever
from ..paths import docs_path, dr_state_path, embeddings_folder_path
from .ann_searcher import ANN_Searcher
from .encoder import Encoder


class DenseRetriever(BaseRetriever):
    def __init__(
        self,
        index_name: str = "new-index",
        model: str = "sentence-transformers/all-MiniLM-L6-v2",
        normalize: bool = True,
        max_length: int = 128,
        use_ann: bool = True,
    ):
        """The Dense Retriever performs [semantic search](https://en.wikipedia.org/wiki/Semantic_search), i.e., it compares vector representations of queries and documents to compute the relevance scores of the latter.

        Args:
            index_name (str, optional): [retriv](https://github.com/AmenRa/retriv) will use `index_name` as the identifier of your index. Defaults to "new-index".

            model (str, optional): defines the encoder model to encode queries and documents into vectors. You can use an [HuggingFace's Transformers](https://huggingface.co/models) pre-trained model by providing its ID or load a local model by providing its path.  In the case of local models, the path must point to the directory containing the data saved with the [`PreTrainedModel.save_pretrained`](https://huggingface.co/docs/transformers/v4.26.1/en/main_classes/model#transformers.PreTrainedModel.save_pretrained) method. Note that the representations are computed with `mean pooling` over the `last_hidden_state`. Defaults to "sentence-transformers/all-MiniLM-L6-v2".

            normalize (bool, optional): whether to L2 normalize the vector representations. Defaults to True.

            max_length (int, optional): texts longer than `max_length` will be automatically truncated. Choose this parameter based on how the employed model was trained or is generally used. Defaults to 128.

            use_ann (bool, optional): whether to use approximate nearest neighbors search. Set it to `False` to use nearest neighbors search without approximation. If you have less than 20k documents in your collection, you probably want to disable approximation. Defaults to True.
        """

        self.index_name = index_name
        self.model = model
        self.normalize = normalize
        self.max_length = max_length
        self.use_ann = use_ann

        self.encoder = Encoder(
            index_name=index_name,
            model=model,
            normalize=normalize,
            max_length=max_length,
        )

        self.ann_searcher = ANN_Searcher(index_name=index_name)

        self.id_mapping = None
        self.doc_count = None
        self.doc_index = None

        self.embeddings = None

    def save(self):
        """Save the state of the retriever to be able to restore it later."""

        state = dict(
            init_args=dict(
                index_name=self.index_name,
                model=self.model,
                normalize=self.normalize,
                max_length=self.max_length,
                use_ann=self.use_ann,
            ),
            id_mapping=self.id_mapping,
            doc_count=self.doc_count,
            embeddings=True if self.embeddings is not None else None,
        )
        np.savez_compressed(dr_state_path(self.index_name), state=state)

    @staticmethod
    def load(index_name: str = "new-index"):
        """Load a retriever and its index.

        Args:
            index_name (str, optional): Name of the index. Defaults to "new-index".

        Returns:
            DenseRetriever: Dense Retriever.
        """

        state = np.load(dr_state_path(index_name), allow_pickle=True)["state"][()]
        dr = DenseRetriever(**state["init_args"])
        dr.initialize_doc_index()
        dr.id_mapping = state["id_mapping"]
        dr.doc_count = state["doc_count"]
        if state["embeddings"]:
            dr.load_embeddings()
        if dr.use_ann:
            dr.ann_searcher = ANN_Searcher.load(index_name)
        return dr

    def load_embeddings(self):
        """Internal usage."""
        path = embeddings_folder_path(self.index_name)
        npy_file_paths = sorted(os.listdir(path))
        self.embeddings = np.concatenate(
            [np.load(path / npy_file_path) for npy_file_path in npy_file_paths]
        )

    def import_embeddings(self, path: str):
        """Internal usage."""
        shutil.copyfile(path, embeddings_folder_path(self.index_name) / "chunk_0.npy")

    def index_aux(
        self,
        embeddings_path: str = None,
        use_gpu: bool = False,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Internal usage."""
        if embeddings_path is not None:
            self.import_embeddings(embeddings_path)
        else:
            self.encoder.change_device("cuda" if use_gpu else "cpu")
            self.encoder.encode_collection(
                path=docs_path(self.index_name),
                batch_size=batch_size,
                callback=callback,
                show_progress=show_progress,
            )
            self.encoder.change_device("cpu")

        if self.use_ann:
            if show_progress:
                print("Building ANN Searcher")
            self.ann_searcher.build()
        else:
            if show_progress:
                print("Loading embeddings...")
            self.load_embeddings()

    def index(
        self,
        collection: Iterable,
        embeddings_path: str = None,
        use_gpu: bool = False,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index a given collection of documents.

        Args:
            collection (Iterable): collection of documents to index.

            embeddings_path (str, optional): in case you want to load pre-computed embeddings, you can provide the path to a `.npy` file. Embeddings must be in the same order as the documents in the collection file. Defaults to None.

            use_gpu (bool, optional): whether to use the GPU for document encoding. Defaults to False.

            batch_size (int, optional): how many documents to encode at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 512.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            DenseRetriever: Dense Retriever
        """

        self.save_collection(collection, callback)
        self.initialize_doc_index()
        self.initialize_id_mapping()
        self.doc_count = len(self.id_mapping)
        self.index_aux(
            embeddings_path=embeddings_path,
            use_gpu=use_gpu,
            batch_size=batch_size,
            callback=callback,
            show_progress=show_progress,
        )
        self.save()
        return self

    def index_file(
        self,
        path: str,
        embeddings_path: str = None,
        use_gpu: bool = False,
        batch_size: int = 512,
        callback: callable = None,
        show_progress: bool = True,
    ):
        """Index the collection contained in a given file.

        Args:
            path (str): path of file containing the collection to index.

            embeddings_path (str, optional): in case you want to load pre-computed embeddings, you can provide the path to a `.npy` file. Embeddings must be in the same order as the documents in the collection file. Defaults to None.

            use_gpu (bool, optional): whether to use the GPU for document encoding. Defaults to False.

            batch_size (int, optional): how many documents to encode at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 512.

            callback (callable, optional): callback to apply before indexing the documents to modify them on the fly if needed. Defaults to None.

            show_progress (bool, optional): whether to show a progress bar for the indexing process. Defaults to True.

        Returns:
            DenseRetriever: Dense Retriever.
        """

        collection = self.collection_generator(path, callback)
        return self.index(
            collection,
            embeddings_path,
            use_gpu,
            batch_size,
            None,
            show_progress,
        )

    def search(
        self,
        query: str,
        return_docs: bool = True,
        cutoff: int = 100,
    ) -> List:
        """Standard search functionality.

        Args:
            query (str): what to search for.

            return_docs (bool, optional): whether to return the texts of the documents. Defaults to True.

            cutoff (int, optional): number of results to return. Defaults to 100.

        Returns:
            List: results.
        """

        encoded_query = self.encoder(query)

        if self.use_ann:
            doc_ids, scores = self.ann_searcher.search(encoded_query, cutoff)
        else:
            if self.embeddings is None:
                self.load_embeddings()
            doc_ids, scores = compute_scores(encoded_query, self.embeddings, cutoff)

        doc_ids = self.map_internal_ids_to_original_ids(doc_ids)

        return (
            self.prepare_results(doc_ids, scores)
            if return_docs
            else dict(zip(doc_ids, scores))
        )

    def msearch(
        self,
        queries: List[Dict[str, str]],
        cutoff: int = 100,
        batch_size: int = 32,
    ) -> Dict:
        """Compute results for multiple queries at once.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

            batch_size (int, optional): how many queries to search at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 32.

        Returns:
            Dict: results.
        """

        q_ids = [x["id"] for x in queries]
        q_texts = [x["text"] for x in queries]
        encoded_queries = self.encoder(q_texts, batch_size, show_progress=False)

        if self.use_ann:
            doc_ids, scores = self.ann_searcher.msearch(encoded_queries, cutoff)
        else:
            if self.embeddings is None:
                self.load_embeddings()
            doc_ids, scores = compute_scores_multi(
                encoded_queries, self.embeddings, cutoff
            )

        doc_ids = [
            self.map_internal_ids_to_original_ids(_doc_ids) for _doc_ids in doc_ids
        ]

        results = {q: dict(zip(doc_ids[i], scores[i])) for i, q in enumerate(q_ids)}

        return {q_id: results[q_id] for q_id in q_ids}

    def bsearch(
        self,
        queries: List[Dict[str, str]],
        cutoff: int = 100,
        batch_size: int = 32,
        show_progress: bool = True,
        qrels: Dict[str, Dict[str, float]] = None,
        path: str = None,
    ):
        """Batch-Search is similar to Multi-Search but automatically generates batches of queries to evaluate and allows dynamic writing of the search results to disk in [JSONl](https://jsonlines.org) format. bsearch is handy for computing results for hundreds of thousands or even millions of queries without hogging your RAM.

        Args:
            queries (List[Dict[str, str]]): what to search for.

            cutoff (int, optional): number of results to return. Defaults to 100.

            batch_size (int, optional): how many queries to search at once. Regulate it if you ran into memory usage issues or want to maximize throughput. Defaults to 32.

            show_progress (bool, optional): whether to show a progress bar for the search process. Defaults to True.

            qrels (Dict[str, Dict[str, float]], optional): query relevance judgements for the queries. Defaults to None.

            path (str, optional): where to save the results. Defaults to None.

        Returns:
            Dict: results.
        """

        batches = [
            queries[i : i + batch_size] for i in range(0, len(queries), batch_size)
        ]

        results = {}

        pbar = tqdm(
            total=len(queries),
            disable=not show_progress,
            desc="Batch search",
            dynamic_ncols=True,
            mininterval=0.5,
        )

        if path is None:
            for batch in batches:
                new_results = self.msearch(
                    queries=batch, cutoff=cutoff, batch_size=len(batch)
                )
                results = {**results, **new_results}
                pbar.update(min(batch_size, len(batch)))
        else:
            path = create_path(path)
            path.parent.mkdir(parents=True, exist_ok=True)

            with open(path, "wb") as f:
                for batch in batches:
                    new_results = self.msearch(queries=batch, cutoff=cutoff)

                    for i, (k, v) in enumerate(new_results.items()):
                        x = {
                            "id": k,
                            "text": batch[i]["text"],
                            "dense_doc_ids": list(v.keys()),
                            "dense_scores": [float(s) for s in list(v.values())],
                        }
                        if qrels is not None:
                            x["rel_doc_ids"] = list(qrels[k].keys())
                            x["rel_scores"] = list(qrels[k].values())
                        f.write(orjson.dumps(x) + "\n".encode())

                    pbar.update(min(batch_size, len(batch)))

        return results


@njit(cache=True)
def compute_scores(query: np.ndarray, docs: np.ndarray, cutoff: int):
    """Internal usage."""

    scores = docs @ query
    indices = np.argsort(-scores)[:cutoff]

    return indices, scores[indices]


@njit(cache=True, parallel=True)
def compute_scores_multi(queries: np.ndarray, docs: np.ndarray, cutoff: int):
    """Internal usage."""

    n = len(queries)
    ids = TypedList([np.empty(1, dtype=np.int64) for _ in range(n)])
    scores = TypedList([np.empty(1, dtype=np.float32) for _ in range(n)])

    for i in prange(len(queries)):
        ids[i], scores[i] = compute_scores(queries[i], docs, cutoff)

    return ids, scores



import re
import string

from unidecode import unidecode


def lowercasing(x: str) -> str:
    return x.lower()


def normalize_ampersand(x: str) -> str:
    return x.replace("&", " and ")


def normalize_diacritics(x: str) -> str:
    return unidecode(x)


def normalize_special_chars(x: str) -> str:
    special_chars_trans = dict(
        [(ord(x), ord(y)) for x, y in zip("‘’´“”–-", "'''\"\"--")]
    )
    return x.translate(special_chars_trans)


def normalize_acronyms(x: str) -> str:
    return re.sub(r"\.(?!(\S[^. ])|\d)", "", x)


def remove_punctuation(x: str) -> str:
    translator = str.maketrans(string.punctuation, " " * len(string.punctuation))
    return x.translate(translator)


def strip_whitespaces(x: str) -> str:
    x = x.strip()

    while "  " in x:
        x = x.replace("  ", " ")

    return x



import numpy as np
from numba import njit


# UNION ------------------------------------------------------------------------
@njit(cache=True)
def union_sorted(a1: np.array, a2: np.array):
    result = np.empty(len(a1) + len(a2), dtype=np.int32)
    i = 0
    j = 0
    k = 0

    while i < len(a1) and j < len(a2):
        if a1[i] < a2[j]:
            result[k] = a1[i]
            i += 1
        elif a1[i] > a2[j]:
            result[k] = a2[j]
            j += 1
        else:  # a1[i] == a2[j]
            result[k] = a1[i]
            i += 1
            j += 1
        k += 1

    result = result[:k]

    if i < len(a1):
        result = np.concatenate((result, a1[i:]))
    elif j < len(a2):
        result = np.concatenate((result, a2[j:]))

    return result


@njit(cache=True)
def union_sorted_multi(arrays):
    if len(arrays) == 1:
        return arrays[0]
    elif len(arrays) == 2:
        return union_sorted(arrays[0], arrays[1])
    else:
        return union_sorted(
            union_sorted_multi(arrays[:2]), union_sorted_multi(arrays[2:])
        )


# INTERSECTION -----------------------------------------------------------------
@njit(cache=True)
def intersect_sorted(a1: np.array, a2: np.array):
    result = np.empty(min(len(a1), len(a2)), dtype=np.int32)
    i = 0
    j = 0
    k = 0

    while i < len(a1) and j < len(a2):
        if a1[i] < a2[j]:
            i += 1
        elif a1[i] > a2[j]:
            j += 1
        else:  # a1[i] == a2[j]
            result[k] = a1[i]
            i += 1
            j += 1
            k += 1

    return result[:k]


@njit(cache=True)
def intersect_sorted_multi(arrays):
    a = arrays[0]

    for i in range(1, len(arrays)):
        a = intersect_sorted(a, arrays[i])

    return a


# DIFFERENCE -------------------------------------------------------------------
@njit(cache=True)
def diff_sorted(a1: np.array, a2: np.array):
    result = np.empty(len(a1), dtype=np.int32)
    i = 0
    j = 0
    k = 0

    while i < len(a1) and j < len(a2):
        if a1[i] < a2[j]:
            result[k] = a1[i]
            i += 1
            k += 1
        elif a1[i] > a2[j]:
            j += 1
        else:  # a1[i] == a2[j]
            i += 1
            j += 1

    result = result[:k]

    if i < len(a1):
        result = np.concatenate((result, a1[i:]))

    return result


#  -----------------------------------------------------------------------------
@njit(cache=True)
def concat1d(X):
    out = np.empty(sum([len(x) for x in X]), dtype=X[0].dtype)

    i = 0
    for x in X:
        for j in range(len(x)):
            out[i] = x[j]
            i = i + 1

    return out


@njit(cache=True)
def get_indices(array, scores):
    n_scores = len(scores)
    min_score = min(scores)
    max_score = max(scores)
    indices = np.full(n_scores, -1, dtype=np.int64)
    counter = 0

    for i in range(len(array)):
        if array[i] >= min_score and array[i] <= max_score:
            for j in range(len(scores)):
                if indices[j] == -1:
                    if scores[j] == array[i]:
                        indices[j] = i
                        counter += 1
                        if len(indices) == counter:
                            return indices
                        break

    return indices


@njit(cache=True)
def unsorted_top_k(array: np.ndarray, k: int):
    top_k_values = np.zeros(k, dtype=np.float32)
    top_k_indices = np.zeros(k, dtype=np.int32)

    min_value = 0.0
    min_value_idx = 0

    for i, value in enumerate(array):
        if value > min_value:
            top_k_values[min_value_idx] = value
            top_k_indices[min_value_idx] = i
            min_value_idx = top_k_values.argmin()
            min_value = top_k_values[min_value_idx]

    return top_k_values, top_k_indices



import math

import faiss
import numpy as np
import psutil
from autofaiss import build_index
from oneliner_utils import read_json

from ..paths import embeddings_folder_path, faiss_index_infos_path, faiss_index_path


def get_ram():
    size_bytes = psutil.virtual_memory().total
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s}GB"


class ANN_Searcher:
    def __init__(self, index_name: str = "new-index"):
        self.index_name = index_name
        self.faiss_index = None
        self.faiss_index_infos = None

    def build(self, use_gpu=False):
        index, index_infos = build_index(
            embeddings=str(embeddings_folder_path(self.index_name)),
            index_path=str(faiss_index_path(self.index_name)),
            index_infos_path=str(faiss_index_infos_path(self.index_name)),
            save_on_disk=True,
            metric_type="ip",
            # max_index_memory_usage="32GB",
            current_memory_available=get_ram(),
            max_index_query_time_ms=10,
            min_nearest_neighbors_to_retrieve=20,
            index_key=None,
            index_param=None,
            use_gpu=use_gpu,
            nb_cores=None,
            make_direct_map=False,
            should_be_memory_mappable=False,
            distributed=None,
            verbose=40,
        )

        self.faiss_index = index
        self.faiss_index_infos = index_infos

    @staticmethod
    def load(index_name: str = "new-index"):
        ann_searcher = ANN_Searcher(index_name)
        ann_searcher.faiss_index = faiss.read_index(str(faiss_index_path(index_name)))
        ann_searcher.faiss_index_infos = read_json(faiss_index_infos_path(index_name))
        return ann_searcher

    def search(self, query: np.ndarray, cutoff: int = 100):
        query = query.reshape(1, len(query))
        ids, scores = self.msearch(query, cutoff)
        return ids[0], scores[0]

    def msearch(self, queries: np.ndarray, cutoff: int = 100):
        scores, ids = self.faiss_index.search(queries, cutoff)
        return ids, scores



from collections import defaultdict
from typing import Dict, Iterable

import numpy as np
from sklearn.feature_extraction.text import CountVectorizer
from tqdm import tqdm


def convert_df_matrix_into_inverted_index(
    df_matrix, vocabulary, show_progress: bool = True
):
    inverted_index = defaultdict(dict)

    for i, term in enumerate(
        tqdm(
            vocabulary,
            disable=not show_progress,
            desc="Building inverted index",
            dynamic_ncols=True,
            mininterval=0.5,
        )
    ):
        inverted_index[term]["doc_ids"] = df_matrix[i].indices
        inverted_index[term]["tfs"] = df_matrix[i].data

    return inverted_index


def build_inverted_index(
    collection: Iterable,
    n_docs: int,
    min_df: int = 1,
    show_progress: bool = True,
) -> Dict:
    vectorizer = CountVectorizer(
        tokenizer=lambda x: x,
        preprocessor=lambda x: x,
        min_df=min_df,
        dtype=np.int16,
        token_pattern=None,
    )

    # [n_docs x n_terms]
    df_matrix = vectorizer.fit_transform(
        tqdm(
            collection,
            total=n_docs,
            disable=not show_progress,
            desc="Building TDF matrix",
            dynamic_ncols=True,
            mininterval=0.5,
        )
    )
    # [n_terms x n_docs]
    df_matrix = df_matrix.transpose().tocsr()
    vocabulary = vectorizer.get_feature_names_out()
    inverted_index = convert_df_matrix_into_inverted_index(
        df_matrix=df_matrix,
        vocabulary=vocabulary,
        show_progress=show_progress,
    )

    doc_lens = np.squeeze(np.asarray(df_matrix.sum(axis=0), dtype=np.float32))
    relative_doc_lens = doc_lens / np.mean(doc_lens, dtype=np.float32)

    return dict(inverted_index), doc_lens, relative_doc_lens



from typing import Tuple

import numba as nb
import numpy as np
from numba import njit, prange
from numba.typed import List as TypedList

from ...utils.numba_utils import (
    intersect_sorted,
    intersect_sorted_multi,
    union_sorted_multi,
    unsorted_top_k,
)


@njit(cache=True)
def bm25(
    b: float,
    k1: float,
    term_doc_freqs: nb.typed.List[np.ndarray],
    doc_ids: nb.typed.List[np.ndarray],
    relative_doc_lens: nb.typed.List[np.ndarray],
    doc_count: int,
    cutoff: int,
    operator: str = "OR",
    subset_doc_ids: np.ndarray = None,
) -> Tuple[np.ndarray]:
    if operator == "AND":
        unique_doc_ids = intersect_sorted_multi(doc_ids)
    elif operator == "OR":
        unique_doc_ids = union_sorted_multi(doc_ids)

    if subset_doc_ids is not None:
        unique_doc_ids = intersect_sorted(unique_doc_ids, subset_doc_ids)

    scores = np.empty(doc_count, dtype=np.float32)
    scores[unique_doc_ids] = 0.0  # Initialize scores

    for i in range(len(term_doc_freqs)):
        indices = doc_ids[i]
        freqs = term_doc_freqs[i]

        df = np.float32(len(indices))
        idf = np.float32(np.log(1.0 + (((doc_count - df) + 0.5) / (df + 0.5))))

        scores[indices] += idf * (
            (freqs * (k1 + 1.0))
            / (freqs + k1 * (1.0 - b + (b * relative_doc_lens[indices])))
        )

    scores = scores[unique_doc_ids]

    if cutoff < len(scores):
        scores, indices = unsorted_top_k(scores, cutoff)
        unique_doc_ids = unique_doc_ids[indices]

    indices = np.argsort(-scores)

    return unique_doc_ids[indices], scores[indices]


@njit(cache=True, parallel=True)
def bm25_multi(
    b: float,
    k1: float,
    term_doc_freqs: nb.typed.List[nb.typed.List[np.ndarray]],
    doc_ids: nb.typed.List[nb.typed.List[np.ndarray]],
    relative_doc_lens: nb.typed.List[np.ndarray],
    doc_count: int,
    cutoff: int,
) -> Tuple[nb.typed.List[np.ndarray]]:
    unique_doc_ids = TypedList([np.empty(1, dtype=np.int32) for _ in doc_ids])
    scores = TypedList([np.empty(1, dtype=np.float32) for _ in doc_ids])

    for i in prange(len(term_doc_freqs)):
        _term_doc_freqs = term_doc_freqs[i]
        _doc_ids = doc_ids[i]

        _unique_doc_ids = union_sorted_multi(_doc_ids)

        _scores = np.empty(doc_count, dtype=np.float32)
        _scores[_unique_doc_ids] = 0.0  # Initialize _scores

        for j in range(len(_term_doc_freqs)):
            indices = _doc_ids[j]
            freqs = _term_doc_freqs[j]

            df = np.float32(len(indices))
            idf = np.float32(np.log(1.0 + (((doc_count - df) + 0.5) / (df + 0.5))))

            _scores[indices] += idf * (
                (freqs * (k1 + 1.0))
                / (freqs + k1 * (1.0 - b + (b * relative_doc_lens[indices])))
            )

        _scores = _scores[_unique_doc_ids]

        if cutoff < len(_scores):
            _scores, indices = unsorted_top_k(_scores, cutoff)
            _unique_doc_ids = _unique_doc_ids[indices]

        indices = np.argsort(_scores)[::-1]

        unique_doc_ids[i] = _unique_doc_ids[indices]
        scores[i] = _scores[indices]

    return unique_doc_ids, scores



from typing import Union

import nltk

from .utils import identity_function

tokenizers_dict = {
    "whitespace": str.split,
    "word": nltk.tokenize.word_tokenize,
    "wordpunct": nltk.tokenize.wordpunct_tokenize,
    "sent": nltk.tokenize.sent_tokenize,
}


def _get_tokenizer(tokenizer: str) -> callable:
    assert tokenizer.lower() in tokenizers_dict, f"Tokenizer {tokenizer} not supported."
    if tokenizer == "punkt":
        nltk.download("punkt", quiet=True)
    return tokenizers_dict[tokenizer.lower()]


def get_tokenizer(tokenizer: Union[str, callable, bool]) -> callable:
    if isinstance(tokenizer, str):
        return _get_tokenizer(tokenizer)
    elif callable(tokenizer):
        return tokenizer
    elif tokenizer is None:
        return identity_function
    else:
        raise (NotImplementedError)



from functools import partial
from typing import Union

import nltk
from krovetzstemmer import Stemmer as KrovetzStemmer
from Stemmer import Stemmer as SnowballStemmer

from .utils import identity_function

stemmers_dict = {
    "krovetz": partial(KrovetzStemmer()),
    "porter": partial(nltk.stem.PorterStemmer().stem),
    "lancaster": partial(nltk.stem.LancasterStemmer().stem),
    "arlstem": partial(nltk.stem.ARLSTem().stem),  # Arabic
    "arlstem2": partial(nltk.stem.ARLSTem2().stem),  # Arabic
    "cistem": partial(nltk.stem.Cistem().stem),  # German
    "isri": partial(nltk.stem.ISRIStemmer().stem),  # Arabic
    "arabic": partial(SnowballStemmer("arabic").stemWord),
    "basque": partial(SnowballStemmer("basque").stemWord),
    "catalan": partial(SnowballStemmer("catalan").stemWord),
    "danish": partial(SnowballStemmer("danish").stemWord),
    "dutch": partial(SnowballStemmer("dutch").stemWord),
    "english": partial(nltk.stem.SnowballStemmer("english").stem),
    "finnish": partial(SnowballStemmer("finnish").stemWord),
    "french": partial(SnowballStemmer("french").stemWord),
    "german": partial(SnowballStemmer("german").stemWord),
    "greek": partial(SnowballStemmer("greek").stemWord),
    "hindi": partial(SnowballStemmer("hindi").stemWord),
    "hungarian": partial(SnowballStemmer("hungarian").stemWord),
    "indonesian": partial(SnowballStemmer("indonesian").stemWord),
    "irish": partial(SnowballStemmer("irish").stemWord),
    "italian": partial(SnowballStemmer("italian").stemWord),
    "lithuanian": partial(SnowballStemmer("lithuanian").stemWord),
    "nepali": partial(SnowballStemmer("nepali").stemWord),
    "norwegian": partial(SnowballStemmer("norwegian").stemWord),
    "portuguese": partial(SnowballStemmer("portuguese").stemWord),
    "romanian": partial(SnowballStemmer("romanian").stemWord),
    "russian": partial(SnowballStemmer("russian").stemWord),
    "spanish": partial(SnowballStemmer("spanish").stemWord),
    "swedish": partial(SnowballStemmer("swedish").stemWord),
    "tamil": partial(SnowballStemmer("tamil").stemWord),
    "turkish": partial(SnowballStemmer("turkish").stemWord),
}


def krovetz_f(x: str) -> str:
    return stemmers_dict["krovetz"](x)


def porter_f(x: str) -> str:
    return stemmers_dict["porter"](x)


def lancaster_f(x: str) -> str:
    return stemmers_dict["lancaster"](x)


def arlstem_f(x: str) -> str:
    return stemmers_dict["arlstem"](x)


def arlstem2_f(x: str) -> str:
    return stemmers_dict["arlstem2"](x)


def cistem_f(x: str) -> str:
    return stemmers_dict["cistem"](x)


def isri_f(x: str) -> str:
    return stemmers_dict["isri"](x)


def arabic_f(x: str) -> str:
    return stemmers_dict["arabic"](x)


def basque_f(x: str) -> str:
    return stemmers_dict["basque"](x)


def catalan_f(x: str) -> str:
    return stemmers_dict["catalan"](x)


def danish_f(x: str) -> str:
    return stemmers_dict["danish"](x)


def dutch_f(x: str) -> str:
    return stemmers_dict["dutch"](x)


def english_f(x: str) -> str:
    return stemmers_dict["english"](x)


def finnish_f(x: str) -> str:
    return stemmers_dict["finnish"](x)


def french_f(x: str) -> str:
    return stemmers_dict["french"](x)


def german_f(x: str) -> str:
    return stemmers_dict["german"](x)


def greek_f(x: str) -> str:
    return stemmers_dict["greek"](x)


def hindi_f(x: str) -> str:
    return stemmers_dict["hindi"](x)


def hungarian_f(x: str) -> str:
    return stemmers_dict["hungarian"](x)


def indonesian_f(x: str) -> str:
    return stemmers_dict["indonesian"](x)


def irish_f(x: str) -> str:
    return stemmers_dict["irish"](x)


def italian_f(x: str) -> str:
    return stemmers_dict["italian"](x)


def lithuanian_f(x: str) -> str:
    return stemmers_dict["lithuanian"](x)


def nepali_f(x: str) -> str:
    return stemmers_dict["nepali"](x)


def norwegian_f(x: str) -> str:
    return stemmers_dict["norwegian"](x)


def portuguese_f(x: str) -> str:
    return stemmers_dict["portuguese"](x)


def romanian_f(x: str) -> str:
    return stemmers_dict["romanian"](x)


def russian_f(x: str) -> str:
    return stemmers_dict["russian"](x)


def spanish_f(x: str) -> str:
    return stemmers_dict["spanish"](x)


def swedish_f(x: str) -> str:
    return stemmers_dict["swedish"](x)


def tamil_f(x: str) -> str:
    return stemmers_dict["tamil"](x)


def turkish_f(x: str) -> str:
    return stemmers_dict["turkish"](x)


stemmers_f_dict = {
    "krovetz": krovetz_f,
    "porter": porter_f,
    "lancaster": lancaster_f,
    "arlstem": arlstem_f,
    "arlstem2": arlstem2_f,
    "cistem": cistem_f,
    "isri": isri_f,
    "arabic": arabic_f,
    "basque": basque_f,
    "catalan": catalan_f,
    "danish": danish_f,
    "dutch": dutch_f,
    "english": english_f,
    "finnish": finnish_f,
    "french": french_f,
    "german": german_f,
    "greek": greek_f,
    "hindi": hindi_f,
    "hungarian": hungarian_f,
    "indonesian": indonesian_f,
    "irish": irish_f,
    "italian": italian_f,
    "lithuanian": lithuanian_f,
    "nepali": nepali_f,
    "norwegian": norwegian_f,
    "portuguese": portuguese_f,
    "romanian": romanian_f,
    "russian": russian_f,
    "spanish": spanish_f,
    "swedish": swedish_f,
    "tamil": tamil_f,
    "turkish": turkish_f,
}


def _get_stemmer(stemmer: str) -> callable:
    assert stemmer.lower() in stemmers_f_dict, f"Stemmer {stemmer} not supported."
    return stemmers_f_dict[stemmer.lower()]


def get_stemmer(stemmer: Union[str, callable, bool]) -> callable:
    if isinstance(stemmer, str):
        return _get_stemmer(stemmer)
    elif callable(stemmer):
        return stemmer
    elif stemmer is None:
        return identity_function
    else:
        raise (NotImplementedError)



__all__ = [
    "lowercasing",
    "normalize_acronyms",
    "normalize_ampersand",
    "normalize_special_chars",
    "remove_punctuation",
    "strip_whitespaces",
    "get_stemmer",
    "get_tokenizer",
    "get_stopwords",
]


from typing import Callable, List, Set

from multipipe import Multipipe

from .normalization import (
    lowercasing,
    normalize_acronyms,
    normalize_ampersand,
    normalize_special_chars,
    remove_punctuation,
    strip_whitespaces,
)
from .stemmer import get_stemmer
from .stopwords import get_stopwords
from .tokenizer import get_tokenizer


def preprocessing(
    x: str,
    tokenizer: Callable,
    stopwords: Set[str],
    stemmer: Callable,
    do_lowercasing: bool,
    do_ampersand_normalization: bool,
    do_special_chars_normalization: bool,
    do_acronyms_normalization: bool,
    do_punctuation_removal: bool,
) -> List[str]:
    if do_lowercasing:
        x = lowercasing(x)
    if do_ampersand_normalization:
        x = normalize_ampersand(x)
    if do_special_chars_normalization:
        x = normalize_special_chars(x)
    if do_acronyms_normalization:
        x = normalize_acronyms(x)

    if tokenizer == str.split and do_punctuation_removal:
        x = remove_punctuation(x)
        x = strip_whitespaces(x)

    x = tokenizer(x)

    if tokenizer != str.split and do_punctuation_removal:
        x = [remove_punctuation(t) for t in x]
        x = [t for t in x if t]

    x = [t for t in x if t not in stopwords]

    return [stemmer(t) for t in x]


def preprocessing_multi(
    tokenizer: callable,
    stopwords: List[str],
    stemmer: callable,
    do_lowercasing: bool,
    do_ampersand_normalization: bool,
    do_special_chars_normalization: bool,
    do_acronyms_normalization: bool,
    do_punctuation_removal: bool,
):
    callables = []

    if do_lowercasing:
        callables.append(lowercasing)
    if do_ampersand_normalization:
        callables.append(normalize_ampersand)
    if do_special_chars_normalization:
        callables.append(normalize_special_chars)
    if do_acronyms_normalization:
        callables.append(normalize_acronyms)
    if tokenizer == str.split and do_punctuation_removal:
        callables.append(remove_punctuation)
        callables.append(strip_whitespaces)

    callables.append(tokenizer)

    if tokenizer != str.split and do_punctuation_removal:

        def rp(x):
            x = [remove_punctuation(t) for t in x]
            return [t for t in x if t]

        callables.append(rp)

    def sw(x):
        return [t for t in x if t not in stopwords]

    callables.append(sw)

    def stem(x):
        return [stemmer(t) for t in x]

    callables.append(stem)

    return Multipipe(callables)



from ranx import Qrels, Run, evaluate, fuse, optimize_fusion


def tune_merger(qrels, runs, metric):
    ranx_qrels = Qrels(qrels)
    ranx_runs = [Run(run) for run in runs]

    best_score = 0.0
    best_config = None

    for norm in ["min-max", "max", "sum"]:
        best_params = optimize_fusion(
            qrels=ranx_qrels,
            runs=ranx_runs,
            norm=norm,
            method="wsum",
            metric=metric,
            show_progress=False,
        )

        combined_run = fuse(
            runs=ranx_runs, norm=norm, method="wsum", params=best_params
        )

        score = evaluate(ranx_qrels, combined_run, metric)
        if score > best_score:
            best_score = score
            best_config = {
                "norm": norm,
                "params": best_params,
            }

    return best_config



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="retriv",
    version="0.2.2",
    author="Elias Bassani",
    author_email="elias.bssn@gmail.com",
    description="retriv: A Python Search Engine for Humans.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AmenRa/retriv",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy",
        "nltk",
        "numba>=0.54.1",
        "tqdm",
        "optuna",
        "krovetzstemmer",
        "pystemmer==2.0.1",
        "unidecode",
        "scikit-learn",
        "ranx",
        "indxr",
        "oneliner_utils",
        "torch",
        "torchvision",
        "torchaudio",
        "transformers[torch]",
        "faiss-cpu",
        "autofaiss",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: General",
    ],
    keywords=[
        "information retrieval",
        "search engine",
        "bm25",
        "numba",
        "sparse retrieval",
        "dense retrieval",
        "hybrid retrieval",
        "neural information retrieval",
    ],
    python_requires=">=3.8",
)



from typing import List, Set, Union

import nltk

supported_languages = {
    "arabic",
    "azerbaijani",
    "basque",
    "bengali",
    "catalan",
    "chinese",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "greek",
    "hebrew",
    "hinglish",
    "hungarian",
    "indonesian",
    "italian",
    "kazakh",
    "nepali",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "slovene",
    "spanish",
    "swedish",
    "tajik",
    "turkish",
}


def _get_stopwords(lang: str) -> List[str]:
    nltk.download("stopwords", quiet=True)
    assert (
        lang.lower() in supported_languages
    ), f"Stop-words for {lang.capitalize()} are not available."
    return nltk.corpus.stopwords.words(lang)


def get_stopwords(sw_list: Union[str, List[str], Set[str], bool]) -> List[str]:
    if isinstance(sw_list, str):
        return _get_stopwords(sw_list)
    elif type(sw_list) is list and all(isinstance(x, str) for x in sw_list):
        return sw_list
    elif type(sw_list) is set:
        return list(sw_list)
    elif sw_list is None:
        return []
    else:
        raise (NotImplementedError)



import os
from pathlib import Path


def base_path():
    p = Path(os.environ.get("RETRIV_BASE_PATH"))
    p.mkdir(parents=True, exist_ok=True)
    return p


def collections_path():
    p = base_path() / "collections"
    p.mkdir(parents=True, exist_ok=True)
    return p


def index_path(index_name: str):
    path = collections_path() / index_name
    path.mkdir(parents=True, exist_ok=True)
    return path


def docs_path(index_name: str):
    return index_path(index_name) / "docs.jsonl"


def sr_state_path(index_name: str):
    return index_path(index_name) / "sr_state.npz"


def fr_state_path(index_name: str):
    return index_path(index_name) / "fr_state.npz"


def embeddings_path(index_name: str):
    return index_path(index_name) / "embeddings.h5"


def embeddings_folder_path(index_name: str):
    path = index_path(index_name) / "embeddings"
    path.mkdir(parents=True, exist_ok=True)
    return path


def faiss_index_path(index_name: str):
    return index_path(index_name) / "faiss.index"


def faiss_index_infos_path(index_name: str):
    return index_path(index_name) / "faiss_index_infos.json"


def dr_state_path(index_name: str):
    return index_path(index_name) / "dr_state.npz"


def hr_state_path(index_name: str):
    return index_path(index_name) / "hr_state.npz"


def encoder_state_path(index_name: str):
    return index_path(index_name) / "encoder_state.json"


def merger_state_path(index_name: str):
    return index_path(index_name) / "merger_state.npz"
